from pathlib import Path
from typing import Protocol, TypeAlias, Callable
from io import StringIO
import itertools
import csv

from PIL import Image
from natsort import natsorted
from loguru import logger

import pcleaner.config as cfg
import pcleaner.structures as st
import pcleaner.helpers as hp
from pcleaner.ocr.ocr_mangaocr import MangaOcr
from pcleaner.ocr.ocr_tesseract import TesseractOcr
import pcleaner.ocr.supported_languages as osl


class OCRModel(Protocol):
    def __call__(self, img_or_path: Image.Image | Path | str, **kwargs) -> str: ...


OCREngineFactory: TypeAlias = Callable[[osl.LanguageCode], OCRModel]


def tesseract_ok(profile: cfg.Profile) -> bool:
    want_tess = profile.preprocessor.ocr_use_tesseract
    has_language_packs: bool = bool(TesseractOcr.langs())
    return not want_tess or (want_tess and has_language_packs)


def ocr_engines():
    return {
        cfg.OCREngine.MANGAOCR: MangaOcr(),
        cfg.OCREngine.TESSERACT: TesseractOcr(),
    }


def get_all_available_langs() -> set[osl.LanguageCode]:
    union_langs = set()
    for ocr in ocr_engines().values():
        union_langs |= ocr.langs()
    return union_langs


def build_ocr_engine_factory(
    tesseract_enabled: bool, ocr_engine_preference: cfg.OCREngine
) -> OCREngineFactory:
    """
    Create a factory function that returns the appropriate OCR engine for a given language.

    :param tesseract_enabled: Whether Tesseract should be considered for use.
    :param ocr_engine_preference: The preferred OCR engine.
    :return: The factory function.
    """
    # want_tess = profile.preprocessor.ocr_use_tesseract
    tess_langs: set[osl.LanguageCode] = TesseractOcr.langs()
    mocr_langs: set[osl.LanguageCode] = MangaOcr.langs()
    if tesseract_enabled and not tess_langs:
        logger.error(
            f"Tesseract OCR is not installed or not found. "
            "Please see the instructions in the README to install Tesseract correctly. Falling back to manga-ocr."
        )

    if not tesseract_enabled:
        ocr_engine_preference = cfg.OCREngine.MANGAOCR

    def closure(lang: osl.LanguageCode) -> OCRModel:
        if ocr_engine_preference == cfg.OCREngine.AUTO:
            # Prefer MangaOCR for the languages it can do, which is only one, but it's
            # much better than Tesseract for it.
            if lang in mocr_langs:
                # The linter is being retarded here.
                # noinspection PyTypeChecker
                return MangaOcr()
            if lang in tess_langs:
                return TesseractOcr(lang)
        elif ocr_engine_preference == cfg.OCREngine.TESSERACT:
            if lang in tess_langs:
                return TesseractOcr(lang)
            logger.error(
                f"Tesseract language pack for '{lang}' is not installed or not found. "
                "Please see the instructions to install Tesseract correctly. Falling back to manga-ocr."
            )
        # Fall back to manga-ocr.
        # noinspection PyTypeChecker
        return MangaOcr()

    return closure


def format_output(
    ocr_analytics: list[st.OCRAnalytic],
    csv_output: bool,
    csv_column_names: tuple[str, str, str, str, str, str],
) -> str:
    """
    Format the output of the OCR process.

    :param ocr_analytics: A list of the OCR analytics per image.
    :param csv_output: If True, output the data in CSV format, otherwise in plain text.
    :param csv_column_names: The (localized) names of the columns in the CSV output:
        filename, startx, starty, endx, endy, text.
    :return: The formatted output.
    """
    # Format of the analytics:
    # number of boxes | sizes of all boxes | sizes of boxes that were OCRed | path to image, text, box coordinates
    # We do not need to show the first three columns, so we simplify the data structure.
    # path_texts_coords: list[tuple[Path, str, st.Box]] = list(
    #     itertools.chain.from_iterable(a.removed_box_data for a in ocr_analytics)
    # )

    # Build tuples of the form (path, text) for the removed texts.
    # This requires adding the path from the analytic to all the texts.
    path_texts_coords: list[tuple[Path, str, st.Box]] = []
    for analytic in ocr_analytics:
        for text, box in analytic.removed_box_data:
            path_texts_coords.append((analytic.path, text, box))

    if path_texts_coords:
        paths, texts, boxes = zip(*path_texts_coords)
        paths = hp.trim_prefix_from_paths(paths)
        path_texts_coords = list(zip(paths, texts, boxes))
        # Sort by path.
        path_texts_coords = natsorted(path_texts_coords, key=lambda x: x[0])

    if csv_output:
        return format_output_csv(path_texts_coords, csv_column_names)
    return format_output_plain(path_texts_coords)


def format_output_csv(
    path_texts_coords: list[tuple[Path, str, st.Box]],
    csv_column_names: tuple[str, str, str, str, str, str],
) -> str:
    buffer = StringIO()
    writer = csv.writer(buffer, quoting=csv.QUOTE_MINIMAL)
    writer.writerow(csv_column_names)

    for path, bubble, box in path_texts_coords:
        if "\n" in bubble:
            logger.warning(f"Detected newline in bubble: {path} {bubble} {box}")
            bubble = bubble.replace("\n", "\\n")
        writer.writerow([path, *box.as_tuple, bubble])

    return buffer.getvalue()


def format_output_plain(path_texts_coords: list[tuple[Path, str, st.Box]]) -> str:
    # Place the file path on it's own line, and only if it's different from the previous one.
    buffer = StringIO()
    current_path = ""
    for path, bubble, _ in path_texts_coords:
        if path != current_path:
            buffer.write(f"\n\n{path}: ")
            current_path = path
        buffer.write(f"\n{bubble}")
        if "\n" in bubble:
            logger.warning(f"Detected newline in bubble: {path} {bubble}")

    return buffer.getvalue()



from enum import IntEnum
from pathlib import Path

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger
from natsort import natsorted

import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.worker_thread as wt
import pcleaner.helpers as hp
import pcleaner.structures as st
from pcleaner.gui.file_table import GUI_PADDING
from pcleaner.gui.ui_generated_files.ui_ImageMatchOverview import Ui_ImageMatchOverview
from pcleaner.helpers import f_plural


class Column(IntEnum):
    IMAGE_INDEX = 0
    SHORT_IMAGE_PATH = 1
    ANALYTICS_PATH = 2


class ImageMatchOverview(Qw.QDialog, Ui_ImageMatchOverview):
    """
    A window to let the user fix a broken image to ocr analytic mapping.
    This happens when loading an old set of OCR results with a different
    set of images.
    """

    images: list[imf.ImageFile]
    mapping: dict[imf.ImageFile, st.OCRAnalytic]
    unmatched_images: list[imf.ImageFile]
    unmatched_analytics: [st.OCRAnalytic]

    def __init__(
        self,
        parent=None,
        mapping: dict[imf.ImageFile, st.OCRAnalytic] = None,
        unmatched_images: list[imf.ImageFile] = None,
        unmatched_analytics: list[st.OCRAnalytic] = None,
    ):
        """
        Init the widget.

        :param parent: The parent widget.
        :param mapping: The mapping of images to OCR analytics.
        :param unmatched_images: The images that don't have a match.
        :param unmatched_analytics: The OCR analytics that don't have a match.
        """
        logger.info(f"Opening the ImageMatchOverview window.")
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.images = list(mapping.keys()) + unmatched_images
        self.images = natsorted(self.images, key=lambda x: x.path)
        self.mapping = mapping
        self.unmatched_images = unmatched_images
        self.unmatched_analytics = natsorted(unmatched_analytics, key=lambda x: x.path)

        self.threadpool = Qc.QThreadPool.globalInstance()

        # Make icons larger so the thumbnails are more visible.
        self.file_table.setIconSize(Qc.QSize(imf.THUMBNAIL_SIZE, imf.THUMBNAIL_SIZE))
        self.file_table.verticalHeader().setDefaultSectionSize(imf.THUMBNAIL_SIZE + GUI_PADDING)

        self.file_table.horizontalHeader().setSectionResizeMode(
            Column.SHORT_IMAGE_PATH, Qw.QHeaderView.Stretch
        )
        self.file_table.horizontalHeader().setSectionResizeMode(
            Column.ANALYTICS_PATH, Qw.QHeaderView.Stretch
        )
        # Hide the index column.
        self.file_table.horizontalHeader().setSectionHidden(Column.IMAGE_INDEX, True)

        self.populate_table()

        self.show_statistics()
        self.init_orphan_warning()

        # Connect the buttons.
        self.pushButton_select_all.clicked.connect(self.select_all)
        self.pushButton_deselect_all.clicked.connect(self.deselect_all)
        self.pushButton_deselect_new.clicked.connect(self.deselect_new)

        self.select_all()
        self.file_table.itemSelectionChanged.connect(self.update_button_states)
        # If there are no images lacking a match, hide the button that removes them.
        if not self.unmatched_images:
            self.pushButton_deselect_new.hide()

        self.buttonBox.button(Qw.QDialogButtonBox.Ok).setText(self.tr("Proceed With Selection"))

    def populate_table(self):

        def make_strikeout_cell(_row: int, _column: int) -> None:
            cell_item = self.file_table.item(_row, _column)
            cell_font = cell_item.font()
            cell_font.setStrikeOut(True)
            cell_item.setFont(cell_font)
            cell_item.setTextAlignment(Qc.Qt.AlignCenter)
            cell_item.setText(" " * 20)

        # First insert the unmatched analytics, since these will be lost.
        for row, analytic in enumerate(self.unmatched_analytics):
            analytic_str = (
                f"{analytic.path} ({len(analytic.removed_box_data)} "
                f"{f_plural(len(analytic.removed_box_data), self.tr('box'), self.tr('boxes'))})"
            )
            self.file_table.appendRow("-1", " " * 20, analytic_str)
            make_strikeout_cell(row, Column.SHORT_IMAGE_PATH)
            # These rows must not be selectable.
            self.file_table.item(row, Column.SHORT_IMAGE_PATH).setFlags(
                self.file_table.item(row, Column.SHORT_IMAGE_PATH).flags() & ~Qc.Qt.ItemIsSelectable
            )
            self.file_table.item(row, Column.ANALYTICS_PATH).setFlags(
                self.file_table.item(row, Column.SHORT_IMAGE_PATH).flags() & ~Qc.Qt.ItemIsSelectable
            )

        # Collect the file paths and map them to their shortened version.
        shortened_paths = hp.trim_prefix_from_paths(list(image.path for image in self.images))

        for row, ((image_index, image_file), short_image_path) in enumerate(
            zip(enumerate(self.images), shortened_paths), start=len(self.unmatched_analytics)
        ):

            analytic_str = ""
            if image_file in self.mapping:
                analytic = self.mapping[image_file]
                analytic_str = (
                    f"{analytic.path} ({len(analytic.removed_box_data)} "
                    f"{f_plural(len(analytic.removed_box_data), self.tr('box'), self.tr('boxes'))})"
                )

            self.file_table.appendRow(f"{image_index}", str(short_image_path), analytic_str)

            if not analytic_str:
                make_strikeout_cell(row, Column.ANALYTICS_PATH)

            self.file_table.item(row, Column.SHORT_IMAGE_PATH).setToolTip(str(image_file.path))

            # Check if the image has loaded yet. If not, use a placeholder icon.
            if not image_file.data_loaded():
                self.file_table.item(row, Column.SHORT_IMAGE_PATH).setIcon(image_file.icon)
            else:
                self.update_row_icon(row, image_file)

        # Begin loading the images in a bit to let the gui update.
        worker = wt.Worker(self.lazy_load_images, no_progress_callback=True)
        worker.signals.error.connect(self.image_dispatch_worker_error)
        self.threadpool.start(worker)

    def update_row_icon(self, row: int, file_obj: imf.ImageFile) -> None:
        """
        Load the image's real thumbnail.

        :param row: The row to update.
        :param file_obj: The file object to use.
        """
        self.file_table.item(row, Column.SHORT_IMAGE_PATH).setIcon(file_obj.thumbnail)

    def show_statistics(self) -> None:
        """
        Compile simple stats and display them above the table.
        """
        plural_images = f_plural(len(self.images), self.tr("image"), self.tr("images"))
        plural_img_unassigned = f_plural(
            len(self.unmatched_images), self.tr("new image"), self.tr("new images")
        )
        stats = self.tr(
            "Matched {num_images} {plural_images} to OCR results. "
            "{num_images_unassigned} {plural_img_unassigned}."
        ).format(
            num_images=len(self.mapping),
            plural_images=plural_images,
            num_images_unassigned=len(self.unmatched_images),
            plural_img_unassigned=plural_img_unassigned,
        )
        self.label_stats.setText(stats)

    def init_orphan_warning(self) -> None:
        """
        If we're dropping analytics, warn the user.
        """
        if not self.unmatched_analytics:
            self.label_warning.hide()
            self.label_warning_icon.hide()
            return

        plural_result = f_plural(
            len(self.unmatched_analytics), self.tr("result"), self.tr("results")
        )
        warning = self.tr(
            "{num_analytics_unassigned} orphaned OCR {plural_result} will be lost."
        ).format(
            num_analytics_unassigned=len(self.unmatched_analytics), plural_result=plural_result
        )
        self.label_warning.setText(warning)

        self.label_warning_icon.setPixmap(Qg.QIcon.fromTheme("dialog-warning").pixmap(16, 16))

    def expand_matched_analytic_paths(self) -> None:
        """
        Expand the paths of the matched analytics to match the full path of the assigned
        image, since the paths from the loaded file are truncated.
        """
        for image, analytic in list(self.mapping.items()):
            analytic_with_expanded_path = st.OCRAnalytic(
                image.path,
                analytic.num_boxes,
                analytic.box_sizes_ocr,
                analytic.box_sizes_removed,
                analytic.removed_box_data,
            )
            self.mapping[image] = analytic_with_expanded_path

    def update_mapping_to_selection(self) -> None:
        """
        When we have finalized a selection, we must provide a blank analytic for each
        new image that didn't have a match.
        Also, unselected images that had a mapping need to be purged.
        """
        selected_indices = self.selected_image_indexes()
        for index, image in enumerate(self.images):
            if index in selected_indices:
                if image in self.mapping:
                    continue
                else:
                    self.mapping[image] = st.OCRAnalytic(image.path, 0, [], [], [])
            else:
                # Remove the mapping.
                self.mapping.pop(image, None)

    def export_final_mapping(self) -> dict[imf.ImageFile, st.OCRAnalytic]:
        """
        Export the final mapping after the user has made their selection.

        :return: The final mapping.
        """
        self.update_mapping_to_selection()
        self.expand_matched_analytic_paths()
        return self.mapping

    def selected_image_indexes(self) -> list[int]:
        """
        Get the indexes of the selected images.

        :return: The indexes of the selected images.
        """
        selected_indexes = []
        for model_index in self.file_table.selectedIndexes():
            if model_index.column() == Column.SHORT_IMAGE_PATH:
                # Read the index from the id column.
                index: int = self.image_index_of_row(model_index.row())
                selected_indexes.append(index)
        return selected_indexes

    def image_index_of_row(self, row: int) -> int | None:
        # Grab the index, but return None if it's -1, since that's the placeholder
        # for analytics without a matching image.
        index = int(self.file_table.item(row, Column.IMAGE_INDEX).text())
        return index if index != -1 else None

    # =========================== Worker Tasks ===========================

    @staticmethod
    def image_loading_task(image: imf.ImageFile) -> Path:
        """
        Thin wrapper to ensure the image object will be returned in the event of an error.

        :param image: The image to load.
        :return: The path to the image for callbacks.
        """
        return image.load_image()

    def lazy_load_images(self) -> None:
        logger.info(f"Dispatching image loading workers")
        # Check for images that haven't loaded their data yet.
        # They will need to have a worker thread scheduled, so the gui doesn't freeze.
        for image in self.images:
            if not (image.loading_queued or image.data_loaded()):
                image.loading_queued = True
                logger.debug(f"Worker Thread loading image {image.path}")
                worker = wt.Worker(self.image_loading_task, image=image, no_progress_callback=True)
                worker.signals.result.connect(self.image_loading_worker_result)
                worker.signals.error.connect(self.image_loading_worker_error)
                self.threadpool.start(worker)

    # ========================= Worker Callbacks =========================

    def image_loading_worker_result(self, file_path: Path) -> None:
        """
        Update the table with the thumbnail.
        """
        # Search for the file in the table.
        for row in range(self.file_table.rowCount()):
            index = int(self.file_table.item(row, Column.IMAGE_INDEX).text())
            if index == -1:
                # Skip the unmatched analytics.
                continue
            image_obj = self.images[index]
            if str(image_obj.path) == str(file_path):
                # Update the table with the thumbnail and image size.
                self.update_row_icon(row, image_obj)
                break
        else:
            logger.warning(f"Worker done. Could not find {file_path} in the table. ignoring.")

    def image_loading_worker_error(self, error: wt.WorkerError) -> None:
        """
        Display an error message in the table.
        """
        # Extract the file object from the WorkerError's *args, being the self argument.
        try:
            file_path = error.kwargs["image"].path
        except (IndexError, AttributeError):
            logger.error(f"Failed to process {error}")
            return
        gu.show_exception(
            self,
            self.tr("Failed to load image"),
            self.tr("Failed to load image {file_path}.").format(file_path=file_path),
            error,
        )

    def image_dispatch_worker_error(self, error: wt.WorkerError) -> None:
        """
        Display an error message in the table.
        """
        gu.show_exception(
            self, self.tr("Dispatch Failed"), self.tr("Failed to dispatch image."), error
        )

    # ============================ Button Slots ============================

    def select_all(self) -> None:
        # We need to exclude the unmatched analytics.
        first_image = self.file_table.model().index(
            len(self.unmatched_analytics), Column.IMAGE_INDEX
        )
        last_image = self.file_table.model().index(
            self.file_table.rowCount() - 1, Column.ANALYTICS_PATH
        )
        self.file_table.selectionModel().select(
            Qc.QItemSelection(first_image, last_image), Qc.QItemSelectionModel.Select
        )
        self.update_button_states()

    def deselect_all(self) -> None:
        self.file_table.clearSelection()
        self.update_button_states()

    def deselect_new(self) -> None:
        """
        Deselect all images that don't have a match.
        """
        for row in range(len(self.unmatched_analytics), self.file_table.rowCount()):
            image_index = self.image_index_of_row(row)
            if self.mapping.get(self.images[image_index]):
                continue
            # Unselect this row.
            model_index = self.file_table.model().index(row, Column.SHORT_IMAGE_PATH)
            self.file_table.selectionModel().select(
                model_index, Qc.QItemSelectionModel.Deselect | Qc.QItemSelectionModel.Rows
            )

        self.update_button_states()

    def update_button_states(self) -> None:
        """
        We only get to show one of the select-/deselect-all buttons at a time.
        We can't let the user click OK if there are no images selected.
        """
        selection = self.selected_image_indexes()
        all_selected = len(selection) == len(self.images)
        any_selected = bool(selection)

        self.pushButton_select_all.setVisible(not all_selected)
        self.pushButton_deselect_all.setVisible(all_selected)

        self.buttonBox.button(Qw.QDialogButtonBox.Ok).setEnabled(any_selected)



import yaml
from pathlib import Path, PosixPath
from importlib import resources

import pcleaner.gui.gui_utils as gu
import pcleaner.gui.ui_generated_files
import pcleaner.data.theme_icons as theme_icons_module
import pcleaner.helpers as hp


def test_get_available_themes():
    """
    Test the get_available_themes function.
    """
    themes = gu.get_available_themes()

    assert themes
    assert set(themes) == {("breeze-dark", "Breeze Dark"), ("breeze", "Breeze Light")}


def test_theme_icon_presence():
    """
    Test that all icons are present in the resources.
    """
    # Read the icon list from the yaml file located at DeepQt/icons/theme_list.yaml
    yaml_path = Path(__file__).parent.parent / "icons" / "theme_list.yaml"
    theme_icons_root = hp.resource_path(theme_icons_module)

    with yaml_path.open() as file:
        data = yaml.safe_load(file)

    theme_directories = data["Theme directories"]
    theme_icons = data["Files"]

    # Perform a depth-first search for each icon in a theme, which means
    # checking for leaf nodes, which are the icon file names.
    # When coming across a directory, the name of that directory is added to the path,
    # then scanned recursively.
    for theme_path in theme_directories:
        # Only use the last directory name of the theme path.
        theme_directory_name = PosixPath(theme_path).name
        for category, sizes in theme_icons.items():
            for size, icons in sizes.items():
                for icon in icons:
                    expected_relative_path = (
                        PosixPath(theme_directory_name) / category / str(size) / icon
                    )

                    for suffix in ["svg", "png"]:
                        full_path = (
                            theme_icons_root
                            / expected_relative_path.with_suffix(f".{suffix}").as_posix()
                        )
                        if full_path.exists():
                            break
                    else:
                        assert (
                            False
                        ), f"Icon {expected_relative_path} not found among the theme_icons data files."


def test_theme_icon_app_presence():
    # Load the source code and inspect it for xdg icon names.
    with resources.as_file(resources.files(pcleaner.gui.ui_generated_files)) as f:
        source_dir = f

    # This only works for the generated code from ui files.
    # General python code would be undecidable.
    expected_icons: set[str] = set()
    for source_file in source_dir.rglob("*.py"):
        with source_file.open() as file:
            for line in file:
                if 'iconThemeName = u"' in line:
                    theme_name = line.split('iconThemeName = u"')[1].split('"')[0]
                    expected_icons.add(theme_name)

    # Ensure they are listed in the yaml icon list file.
    yaml_path = Path(__file__).parent.parent / "icons" / "theme_list.yaml"
    with yaml_path.open() as file:
        data = yaml.safe_load(file)

    theme_icons = data["Files"]
    icon_set = set()
    for category, sizes in theme_icons.items():
        for size, icons in sizes.items():
            for icon in icons:
                icon_set.add(icon)

    missing = expected_icons - icon_set
    assert not missing



from pathlib import Path
import shutil

from PIL import Image
from loguru import logger

import pcleaner.image_ops as ops
import pcleaner.structures as st
import pcleaner.output_structures as ost


def denoise_page(d_data: st.DenoiserData) -> st.DenoiseAnalytic:
    """
    Load the MaskData from the json file and perform the denoising process.

    :param d_data: All the data needed for the denoising process.
    :return: Analytics.
    """
    # Load all the cached data.
    mask_data = st.MaskData.from_json(d_data.json_path.read_text(encoding="utf-8"))
    mask_image = Image.open(mask_data.mask_path)

    original_path = Path(mask_data.original_path)
    path_gen = ost.OutputPathGenerator(original_path, d_data.cache_dir, d_data.json_path)

    def save_mask(img, path: Path) -> None:
        img.save(path)

    cleaned_image = Image.open(mask_data.original_path)
    # Check if we're dealing with a 1-bit image.
    # Denoising these is pointless, as they have no noise.
    if cleaned_image.mode == "1":
        logger.info(f"Skipping pointless denoising for 1-bit image: {original_path}")
        # We can simply copy the cleaned output from the masker step.
        shutil.copyfile(path_gen.clean, path_gen.clean_denoised)
        # The noise mask is just a blank image with with the same size as the original.
        blank_noise_mask = Image.new("LA", cleaned_image.size, (0, 0))
        save_mask(blank_noise_mask, path_gen.noise_mask)
        return st.DenoiseAnalytic(tuple(), original_path)

    # Scale the mask to the original image size, if needed.
    cleaned_image = cleaned_image.convert("RGB")
    scale_up_factor = 1.0
    if cleaned_image.size != mask_image.size:
        scale_up_factor = cleaned_image.size[0] / mask_image.size[0]
        mask_image = mask_image.resize(cleaned_image.size, resample=Image.NEAREST)

    cleaned_image.paste(mask_image, (0, 0), mask_image)
    original_path: Path = mask_data.original_path

    # Alias.
    d_conf = d_data.denoiser_config

    # Filter for the min deviation to consider for denoising.
    boxes_to_denoise: list[st.Box] = [
        box
        for box, deviation, failed, _ in mask_data.boxes_with_stats
        if not failed and deviation > d_conf.noise_min_standard_deviation
    ]

    noise_masks_with_coords: list[tuple[Image.Image, tuple[int, int]]] = [
        ops.generate_noise_mask(cleaned_image, mask_image, box, d_conf, scale_up_factor)
        for box in boxes_to_denoise
    ]

    if noise_masks_with_coords:
        combined_noise_mask = ops.combine_noise_masks(cleaned_image.size, noise_masks_with_coords)
        cleaned_image.paste(combined_noise_mask, (0, 0), combined_noise_mask)
    else:
        # noinspection PyTypeChecker
        combined_noise_mask = Image.new("LA", cleaned_image.size, (0, 0))

    save_mask(combined_noise_mask, path_gen.noise_mask)
    save_mask(cleaned_image, path_gen.clean_denoised)

    # Package the analytics. We're only interested in the std deviations.
    return st.DenoiseAnalytic(
        tuple(deviation for _, deviation, _, _ in mask_data.boxes_with_stats), original_path
    )



import math
from enum import IntEnum
from functools import wraps, partial

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from PIL import Image
from PySide6.QtCore import Slot
from loguru import logger
from natsort import natsorted

import pcleaner.gui.gui_utils as gu
import pcleaner.gui.image_file as imf
import pcleaner.gui.state_saver as ss
import pcleaner.gui.structures as gst
import pcleaner.gui.worker_thread as wt
import pcleaner.ocr.ocr as ocr
import pcleaner.ocr.supported_languages as osl
import pcleaner.output_structures as ost
import pcleaner.structures as st
from pcleaner.gui.ui_generated_files.ui_OcrReview import Ui_OcrReview
from pcleaner.helpers import tr, f_plural


# The maximum size, will be smaller on one side if the image is not square.
THUMBNAIL_SIZE = 180


# The way the cellUpdate signal works, which we need to know when a cell was manually edited,
# is that it also triggers on all other changes too... But we need to ignore those
# due to inconsistent intermediate states causing exception.
def suppress_cell_update_handling(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        self.awaiting_user_input = False
        result = method(self, *args, **kwargs)
        self.awaiting_user_input = True
        return result

    return wrapper


class ViewMode(IntEnum):
    WithBoxes = 0
    Original = 1


BUBBLE_STATUS_COLORS = {
    st.OCRStatus.Normal: Qg.QColor(0, 128, 0, 255),
    st.OCRStatus.Removed: Qg.QColor(128, 0, 0, 255),
    st.OCRStatus.EditedRemoved: Qg.QColor(128, 0, 0, 255),
    st.OCRStatus.Edited: Qg.QColor(0, 0, 128, 255),
    st.OCRStatus.New: Qg.QColor(128, 64, 0, 255),
}


class OcrReviewWindow(Qw.QDialog, Ui_OcrReview):
    """
    A window to display the process results.
    """

    images: list[imf.ImageFile]
    # These are the in the raw analytic format.
    ocr_analytics: list[st.OCRAnalytic]
    # These use the mutable format.
    ocr_results: list[list[st.OCRResult]]
    # When editing old data, we shouldn't try to use the generated box images,
    # as they might not exist or be wrong.
    editing_old_data: bool

    min_thumbnail_size: int
    max_thumbnail_size: int

    first_load: bool

    ocr_engine_factory: gst.Shared[ocr.OCREngineFactory]
    theme_is_dark: gst.Shared[bool]

    # When True, handle changes to cells as user input.
    awaiting_user_input: bool

    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
        images: list[imf.ImageFile] = None,
        ocr_analytics: list[st.OCRAnalytic] = None,
        editing_old_data: bool = False,
        ocr_engine_factory: gst.Shared[ocr.OCREngineFactory] = None,
        theme_is_dark: gst.Shared[bool] = None,
    ):
        """
        Init the widget.

        :param ocr_engine_factory:
        :param parent: The parent widget.
        :param images: The images to display.
        :param ocr_analytics: The OCR results to display.
        :param editing_old_data: When true, don't attempt to load generated box outputs.
        :param theme_is_dark: The shared theme state.
        """
        logger.info(f"Opening OCR Review Window for {len(images)} outputs.")
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        # Special OutputReview overrides.
        self.target_output = ost.Output.initial_boxes
        self.confirm_closing = True

        self.images = images
        self.ocr_analytics = ocr_analytics
        self.editing_old_data = editing_old_data
        self.ocr_engine_factory = ocr_engine_factory
        self.theme_is_dark = theme_is_dark

        self.sort_images_by_path()

        self.awaiting_user_input = False

        if len(self.images) != len(self.ocr_analytics):
            logger.error("The number of images and OCR results don't match.")

        self.ocr_results = st.convert_ocr_analytics_to_results(self.ocr_analytics)

        self.first_load = True

        # Set the table to only allow editing in the text column.
        self.tableWidget_ocr.setEditableColumns([1])
        self.tableWidget_ocr.setItemDelegate(WrapTextDelegate())
        self.tableWidget_ocr.resized.connect(self.adjust_row_heights)

        self.calculate_thumbnail_size()
        self.init_arrow_buttons()
        self.init_image_list()
        self.image_list.currentItemChanged.connect(self.handle_image_change)
        # Select the first image to start with.
        self.comboBox_view_mode.currentIndexChanged.connect(self.update_image_boxes)
        self.tableWidget_ocr.currentRowChanged.connect(self.update_image_boxes)
        Qc.QTimer.singleShot(0, partial(self.image_list.setCurrentRow, 0))
        # Override Qt's dynamic scroll speed with a fixed, standard value.
        self.image_list.verticalScrollBar().setSingleStep(120)

        # Let the user select the bubble by clicking it too.
        self.image_viewer.bubble_clicked.connect(self.handle_bubble_clicked)

        self.pushButton_done.clicked.connect(self.close)

        # Connect the zoom buttons to the image viewer.
        self.pushButton_zoom_in.clicked.connect(self.image_viewer.zoom_in)
        self.pushButton_zoom_out.clicked.connect(self.image_viewer.zoom_out)
        self.pushButton_zoom_fit.clicked.connect(self.image_viewer.zoom_fit)
        self.pushButton_zoom_reset.clicked.connect(self.image_viewer.zoom_reset)

        # Connect the table buttons.
        self.pushButton_new.toggled.connect(self.new_bubble)
        self.pushButton_delete.clicked.connect(self.delete_bubble)
        self.pushButton_reset.clicked.connect(self.reset_single_bubble)
        self.pushButton_reset_all.clicked.connect(self.reset_all_bubbles)
        self.pushButton_row_up.clicked.connect(self.move_row_up)
        self.pushButton_row_down.clicked.connect(self.move_row_down)
        self.pushButton_undelete.clicked.connect(self.delete_bubble)
        self.tableWidget_ocr.currentRowChanged.connect(self.update_button_availability)
        self.tableWidget_ocr.cellChanged.connect(self.handle_table_edited)

        # Callback for new bubble creation.
        self.image_viewer.new_bubble.connect(self.add_new_bubble)

        # Make side splitter 50/50.
        side_splitter_width = self.splitter_side.width()
        self.splitter_side.setSizes([side_splitter_width // 2, side_splitter_width // 2])

        self.load_custom_icons()
        self.update_button_availability()

        # Set the new button color to match what would be assigned to new buttons.
        self.image_viewer.set_new_bubble_color(BUBBLE_STATUS_COLORS[st.OCRStatus.New])

        self.init_shortcuts()

        self.load_ocr_options()

        self.state_saver = ss.StateSaver("ocr_review")
        self.init_state_saver()
        self.state_saver.restore()

    def init_shortcuts(self) -> None:
        def set_button_shortcut(button: Qw.QPushButton, key_sequence: Qg.QKeySequence) -> None:
            button.setShortcut(key_sequence)
            button.setToolTip(button.toolTip() + f" ({button.shortcut().toString()})")

        set_button_shortcut(self.pushButton_new, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_N))
        set_button_shortcut(self.pushButton_delete, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_D))
        set_button_shortcut(
            self.pushButton_undelete, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.SHIFT + Qc.Qt.Key_D)
        )
        set_button_shortcut(self.pushButton_reset, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_R))
        set_button_shortcut(self.pushButton_prev, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_Left))
        set_button_shortcut(self.pushButton_next, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_Right))
        set_button_shortcut(self.pushButton_row_up, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_Up))
        set_button_shortcut(self.pushButton_row_down, Qg.QKeySequence(Qc.Qt.CTRL + Qc.Qt.Key_Down))

    def get_final_ocr_analytics(self) -> list[st.OCRAnalytic]:
        """
        Convert the mutable OCR results back to the original format.

        :return: The OCR analytics.
        """
        return st.convert_ocr_results_to_analytics(self.ocr_results)

    def load_custom_icons(self) -> None:
        # Load the custom new_bubble icon.
        theme = "dark" if self.theme_is_dark.get() else "light"

        icon_new = gu.load_custom_icon("new_bubble", theme)
        icon_undelete = gu.load_custom_icon("trash_restore", theme)

        # Check if it loaded correctly.
        if icon_new.isNull():
            logger.error("Failed to load icon new_bubble.")
            return
        if icon_undelete.isNull():
            logger.error("Failed to load icon trash_restore.")
            return

        self.pushButton_new.setIcon(icon_new)
        self.pushButton_new.setText("")
        self.pushButton_undelete.setIcon(icon_undelete)
        self.pushButton_undelete.setText("")

    def switch_to_image(self, image: imf.ImageFile) -> None:
        """
        Show the image in the button in the image view.

        :param image: The image to show.
        """
        original_path = image.path

        self.label_file_name.setText(str(original_path))
        self.label_file_name.setToolTip(str(original_path))
        self.label_file_name.setElideMode(Qc.Qt.ElideLeft)

        self.image_viewer.set_image(original_path)

        self.load_ocr_results(self.ocr_results[self.images.index(image)])

        # Select the first row (if any) to allow for easy keyboard navigation.
        if self.tableWidget_ocr.rowCount() > 0:
            self.tableWidget_ocr.setCurrentCell(0, 1)
            self.tableWidget_ocr.focusWidget()

        # Set a delayed zoom to fit.
        if self.first_load:
            Qc.QTimer.singleShot(0, self.image_viewer.zoom_fit)
            self.first_load = False

    @suppress_cell_update_handling
    def load_ocr_results(self, ocr_results: list[st.OCRResult]) -> None:
        """
        Load the OCR results into the text fields.

        :param ocr_results: The OCR results to load for this image.
        """
        self.tableWidget_ocr.clearAll()

        # This index is for the table to keep track of the index into the specific OCR result list.
        # It isn't the original index of the box as shown in the image.
        for index, ocr_result in enumerate(ocr_results):
            self.tableWidget_ocr.appendRow(
                str(ocr_result.label),
                ocr_result.text,
            )
            # If the status is edited, make it italic, removed is strikeout, and
            # editedremoved is both.
            cell_font = self.tableWidget_ocr.item(index, 1).font()
            if ocr_result.status == st.OCRStatus.Edited:
                cell_font.setItalic(True)
            elif ocr_result.status == st.OCRStatus.Removed:
                cell_font.setStrikeOut(True)
            elif ocr_result.status == st.OCRStatus.EditedRemoved:
                cell_font.setItalic(True)
                cell_font.setStrikeOut(True)
            self.tableWidget_ocr.item(index, 1).setFont(cell_font)

        self.update_image_boxes()
        self.adjust_row_heights()

    def adjust_row_heights(self):
        for row in range(self.tableWidget_ocr.rowCount()):
            self.tableWidget_ocr.resizeRowToContents(row)

    def update_image_boxes(self) -> None:
        """
        Update the image viewer with the current OCR results.
        """
        if self.view_mode() == ViewMode.Original or not self.ocr_results:
            self.image_viewer.clear_bubbles()
            return

        ocr_results = self.current_image_ocr_results()

        rects: list[Qc.QRect] = [
            Qc.QRect(*ocr_result.box.as_tuple_xywh) for ocr_result in ocr_results
        ]
        colors: list[Qg.QColor] = [
            BUBBLE_STATUS_COLORS[ocr_result.status] for ocr_result in ocr_results
        ]
        labels: list[str] = [str(ocr_result.label) for ocr_result in ocr_results]
        strokes: list[Qc.Qt.PenStyle] = list(
            map(
                lambda r: (
                    Qc.Qt.SolidLine
                    if r.status not in (st.OCRStatus.Removed, st.OCRStatus.EditedRemoved)
                    else Qc.Qt.DashLine
                ),
                ocr_results,
            )
        )
        # assert self.tableWidget_ocr.rowCount() == len(ocr_results)
        # assert self.tableWidget_ocr.rowCount() == len(rects)
        # assert self.tableWidget_ocr.rowCount() == len(colors)
        # assert self.tableWidget_ocr.rowCount() == len(labels)
        # assert self.tableWidget_ocr.rowCount() == len(strokes)

        # Set the selected bubble to have the highlight color.
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row != -1:
            colors[selected_row] = self.palette().highlight().color()
        self.image_viewer.set_bubbles(rects, colors, labels, strokes)

    def view_mode(self) -> ViewMode:
        """
        Get the current view mode.

        :return: The current view mode.
        """
        return ViewMode(self.comboBox_view_mode.currentIndex())

    def current_image_ocr_results(self) -> list[st.OCRResult]:
        """
        Get the OCR results for the currently selected image.

        :return: The OCR results.
        """
        return self.ocr_results[self.image_list.currentRow()]

    @Slot(int)
    def handle_bubble_clicked(self, index: int) -> None:
        self.tableWidget_ocr.setCurrentCell(index, 1)

    # OCR Engine stuff =============================================================================

    def load_ocr_options(self) -> None:
        """
        We can either disable auto OCR or choose what language we want it to detect.
        """
        self.comboBox_ocr_language.addTextItemLinkedData(self.tr("No OCR"), None)
        code_and_name = osl.language_code_name_sorted(
            include_detect=False, pin_important=True, translate=tr
        )
        # Filter out the languages we can't use anyway.
        available_lang_codes = ocr.get_all_available_langs()
        for code, lang in filter(lambda x: x[0] in available_lang_codes, code_and_name):
            self.comboBox_ocr_language.addTextItemLinkedData(lang, code)
        # Start at the second item, ie. the first actual language.
        self.comboBox_ocr_language.setCurrentIndex(1)

    def get_ocr_engine(self) -> ocr.OCRModel | None:
        """
        Get the current OCR model to use.

        :return: The OCR model to use, or None if no OCR should be done.
        """
        selected_index = self.comboBox_ocr_language.currentIndex()
        if selected_index == 0:
            return None
        return self.ocr_engine_factory.get()(self.comboBox_ocr_language.currentLinkedData())

    def start_ocr(self, box: st.Box, row: int) -> None:
        """
        Start a worker thread to perform OCR on the given box.
        It's typically fast enough to not be noticeable if it were blocking,
        then again I don't know what kind of potato you're running this on.

        :param box: The box to perform OCR on.
        :param row: The row it was added on.
        """
        current_image_index = self.image_list.currentRow()
        ocr_worker = wt.Worker(
            self.perform_ocr, box, current_image_index, row, no_progress_callback=True
        )
        ocr_worker.signals.error.connect(self.ocr_worker_error)
        ocr_worker.signals.result.connect(self.ocr_worker_result)
        ocr_worker.signals.finished.connect(self.ocr_worker_finished)
        Qc.QThreadPool.globalInstance().start(ocr_worker)

        Qw.QApplication.setOverrideCursor(Qc.Qt.WaitCursor)

    def perform_ocr(self, box: st.Box, image_index: int, row: int) -> tuple[str, int, int]:
        """
        Grab the image file directly and crop that.

        :param box: The box to crop.
        :param image_index: The index of the image this is for.
            Important to know if the user has switched images.
        :param row: The row in the table to update.
        :return: The OCR result.
        """
        model = self.get_ocr_engine()
        original_image = self.image_viewer.image_item.pixmap()
        # Crop the image to the bubble.
        bubble_image = original_image.copy(*box.as_tuple_xywh)
        bubble_pil_image = Image.fromqpixmap(bubble_image)
        return model(bubble_pil_image), image_index, row

    def ocr_worker_error(self, error: wt.WorkerError) -> None:
        logger.error(f"OCR worker thread error: {error}")
        gu.show_exception(self, self.tr("OCR Error"), self.tr("Encountered error:"), error)

    @suppress_cell_update_handling
    def ocr_worker_result(self, result: tuple[str, int, int]) -> None:
        text, image_index, row = result

        ocr_results = self.ocr_results[image_index]
        try:
            ocr_results[row].text = text
        except IndexError:
            logger.error("OCR result index out of range. User likely deleted the bubble.")

        self.load_ocr_results(ocr_results)

        # Focus the newly created bubble, if still on this image.
        if image_index == self.image_list.currentRow():
            self.tableWidget_ocr.setCurrentCell(row, 1)
            self.tableWidget_ocr.focusWidget()

    def ocr_worker_finished(self) -> None:
        # Reset the cursor and enable the window.
        Qw.QApplication.restoreOverrideCursor()

    # Table manipulation functions =================================================================

    def update_button_availability(self) -> None:
        """
        Check which buttons should be enabled.
        """
        selected_row = self.tableWidget_ocr.currentRow()
        self.pushButton_row_up.setEnabled(selected_row > 0)
        self.pushButton_row_down.setEnabled(
            selected_row != -1 and selected_row < self.tableWidget_ocr.rowCount() - 1
        )
        # For resetting a single bubble, both the index needs to
        # be valid and the bubble must be edited.
        if selected_row == -1:
            self.pushButton_reset.setEnabled(False)
        else:
            ocr_results = self.current_image_ocr_results()
            self.pushButton_reset.setEnabled(
                ocr_results[selected_row].status
                in (st.OCRStatus.Edited, st.OCRStatus.EditedRemoved, st.OCRStatus.Removed)
            )
        # For delete there must be a selection and it can't be deleted already.
        self.pushButton_undelete.hide()
        if selected_row == -1:
            self.pushButton_delete.setEnabled(False)
            self.pushButton_delete.show()
            self.pushButton_undelete.hide()
        else:
            ocr_results = self.current_image_ocr_results()
            self.pushButton_delete.setEnabled(True)
            if ocr_results[selected_row].status in (
                st.OCRStatus.Removed,
                st.OCRStatus.EditedRemoved,
            ):
                self.pushButton_delete.hide()
                self.pushButton_undelete.show()
            else:
                self.pushButton_delete.show()
                self.pushButton_undelete.hide()

    @suppress_cell_update_handling
    def delete_bubble(self) -> None:
        """
        Remove the currently selected bubble.
        If it was a new bubble, remove it entirely.
        Otherwise, mark it as removed.
        But if already removed, undelete it.
        """
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row == -1:
            return

        ocr_results = self.current_image_ocr_results()
        ocr_result = ocr_results[selected_row]

        if ocr_result.status == st.OCRStatus.New:
            ocr_results.pop(selected_row)
        elif ocr_result.status == st.OCRStatus.Removed:
            ocr_result.status = st.OCRStatus.Normal
        elif ocr_result.status == st.OCRStatus.EditedRemoved:
            ocr_result.status = st.OCRStatus.Edited
        elif ocr_result.status == st.OCRStatus.Edited:
            ocr_result.status = st.OCRStatus.EditedRemoved
        else:
            ocr_result.status = st.OCRStatus.Removed

        self.load_ocr_results(ocr_results)

    @suppress_cell_update_handling
    def reset_single_bubble(self) -> None:
        """
        Reset only the currently selected bubble.
        Go by the original label to find the correct bubble.
        """
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row == -1:
            return
        # Check that this is an edited bubble.
        ocr_results = self.current_image_ocr_results()
        if ocr_results[selected_row].status not in (
            st.OCRStatus.Edited,
            st.OCRStatus.EditedRemoved,
            st.OCRStatus.Removed,
        ):
            logger.error("Tried to reset a non-edited bubble.")
            return

        image_index = self.image_list.currentRow()
        original_analytic = self.ocr_analytics[image_index]
        # Wrap it in a list due to the way the helper is written, then unpack the result.
        original_results = st.convert_ocr_analytics_to_results([original_analytic])[0]

        ocr_results = self.ocr_results[image_index]
        # Figure out what the original row was by matching the label.
        original_row = next(
            i
            for i, ocr_result in enumerate(original_results)
            if ocr_result.label == ocr_results[selected_row].label
        )
        ocr_results[selected_row] = original_results[original_row]

        self.load_ocr_results(ocr_results)

    @suppress_cell_update_handling
    def reset_all_bubbles(self) -> None:
        """
        Reset all bubbles to their original state.
        """
        # Confirm first.
        if (
            gu.show_question(
                self,
                self.tr("Reset Bubbles"),
                self.tr("Are you sure you want to reset all boxes for this image?"),
            )
            == Qw.QMessageBox.Cancel
        ):
            return
        # Pull the original OCR results from the analytics.
        image_index = self.image_list.currentRow()
        original_analytic = self.ocr_analytics[image_index]
        # Wrap it in a list due to the way the helper is written, then unpack the result.
        original_results = st.convert_ocr_analytics_to_results([original_analytic])[0]

        self.ocr_results[image_index] = original_results
        self.load_ocr_results(original_results)

    @suppress_cell_update_handling
    def move_row_up(self) -> None:
        """
        Move the currently selected row up.
        """
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row == -1 or selected_row == 0:
            return

        ocr_results = self.current_image_ocr_results()
        ocr_results.insert(selected_row - 1, ocr_results.pop(selected_row))

        self.load_ocr_results(ocr_results)
        self.tableWidget_ocr.setCurrentCell(selected_row - 1, 1)

    @suppress_cell_update_handling
    def move_row_down(self) -> None:
        """
        Move the currently selected row down.
        """
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row == -1 or selected_row == self.tableWidget_ocr.rowCount() - 1:
            return

        ocr_results = self.current_image_ocr_results()
        ocr_results.insert(selected_row + 1, ocr_results.pop(selected_row))

        self.load_ocr_results(ocr_results)
        self.tableWidget_ocr.setCurrentCell(selected_row + 1, 1)

    def new_bubble(self) -> None:
        """
        Switch the image viewer to bubble mode.
        The actual bubble creation is done once the callback comes in on
        BubbleImageViewer.new_bubble.
        """
        self.image_viewer.set_allow_drawing_bubble(self.pushButton_new.isChecked())

    @suppress_cell_update_handling
    def add_new_bubble(self, rect: Qc.QRect) -> None:
        # Disable new bubble mode.
        self.pushButton_new.setChecked(False)
        self.new_bubble()

        # Create regular xyxy coordinates for a box.
        box = st.Box(*rect.getCoords())

        # Discard if the bubble is too small.
        MIN_BUBBLE_SIZE = 400
        if box.area < MIN_BUBBLE_SIZE:
            logger.debug("Discarding bubble due to size.")
            return

        # Find the lowest free bubble index.
        first_free_index = 1
        used_indices = [
            int(ocr_result.label.split()[-1])
            for ocr_result in self.current_image_ocr_results()
            if ocr_result.status == st.OCRStatus.New
        ]
        while first_free_index in used_indices:
            first_free_index += 1

        new_bubble_label = self.tr("New") + f" {first_free_index}"

        # Add the new bubble to the current image.
        ocr_results = self.current_image_ocr_results()
        image_path = self.image_viewer.loaded_image_path
        # Insert it after the currently selected box, if any, otherwise at the end.
        selected_row = self.tableWidget_ocr.currentRow()
        if selected_row == -1:
            new_row = self.tableWidget_ocr.rowCount()
            ocr_results.append(
                st.OCRResult(image_path, "", box, new_bubble_label, st.OCRStatus.New)
            )
        else:
            new_row = selected_row + 1
            ocr_results.insert(
                new_row,
                st.OCRResult(image_path, "", box, new_bubble_label, st.OCRStatus.New),
            )

        self.load_ocr_results(ocr_results)

        if self.get_ocr_engine() is not None:
            # Quick sanity check. We might not want to OCR more than 20% of the image.
            # This is a rough estimate, but it should be good enough.
            image_area = (
                self.image_viewer.image_dimensions[0] * self.image_viewer.image_dimensions[1]
            )
            bubble_area = box.area
            if bubble_area > 0.2 * image_area:
                if (
                    gu.show_question(
                        self,
                        self.tr("OCR Warning"),
                        self.tr("The bubble is very large. Are you sure you want to OCR it?"),
                    )
                    == Qw.QMessageBox.Cancel
                ):
                    return
            self.start_ocr(box, new_row)
        else:
            # Focus the newly created bubble.
            self.tableWidget_ocr.setCurrentCell(new_row, 1)
            self.tableWidget_ocr.focusWidget()

    @Slot(int, int)
    def handle_table_edited(self, row: int, col: int) -> None:
        """
        Store the edited text in the OCR results.

        :param row: The row of the edited cell.
        :param col: The column of the edited cell, must be 1, the label may not be changed.
        """
        if not self.awaiting_user_input:
            return

        if col != 1:
            logger.error("Only the text column should be editable.")
            return

        ocr_results = self.current_image_ocr_results()
        ocr_results[row].text = self.tableWidget_ocr.currentText(1)
        # If the cell was normal, mark it as edited.
        if ocr_results[row].status == st.OCRStatus.Normal:
            ocr_results[row].status = st.OCRStatus.Edited
        if ocr_results[row].status == st.OCRStatus.Removed:
            ocr_results[row].status = st.OCRStatus.EditedRemoved

        current_row = self.tableWidget_ocr.currentRow()
        self.load_ocr_results(ocr_results)
        self.tableWidget_ocr.selectRow(current_row)

    # Copied shit from OutputReview ================================================================

    def sort_images_by_path(self) -> None:
        """
        Sort the images by their file path using natsort.
        This is necessary because the parallel batch processing will not preserve the order when
        many pictures are processed at once.
        """
        # self.images = natsorted(self.images, key=lambda x: x.path)
        combined = list(zip(self.images, self.ocr_analytics))
        sorted_combined = natsorted(combined, key=lambda x: x[0].path)
        images, ocr_analytics = zip(*sorted_combined)
        self.images = list(images)
        self.ocr_analytics = list(ocr_analytics)

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        if self.confirm_closing:
            if (
                gu.show_question(
                    self,
                    self.tr("Finish Review"),
                    self.tr("Are you sure you want to finish the review?"),
                )
                == Qw.QMessageBox.Cancel
            ):
                event.ignore()
                return

        self.state_saver.save()
        event.accept()

    def calculate_thumbnail_size(self) -> None:
        """
        Use the current monitor's resolution to calculate the thumbnail size.
        The min value is 1% of the screen width, the max value is 100%.
        """
        screen = Qw.QApplication.primaryScreen()
        screen_size = screen.size()
        screen_width = screen_size.width()
        self.min_thumbnail_size = int(screen_width * 0.01)
        self.max_thumbnail_size = screen_width

    def to_log_scale(self, value: int) -> int:
        """
        Convert a linear value to a logarithmic scale.
        """
        min_log = math.log(self.min_thumbnail_size)
        max_log = math.log(self.max_thumbnail_size)
        scale = (max_log - min_log) / (
            self.horizontalSlider_icon_size.maximum() - self.horizontalSlider_icon_size.minimum()
        )
        log_value = min_log + (value - self.horizontalSlider_icon_size.minimum()) * scale
        return int(math.exp(log_value))

    def from_log_scale(self, value: int) -> int:
        """
        Convert a logarithmic value to a linear scale.
        """
        min_log = math.log(self.min_thumbnail_size)
        max_log = math.log(self.max_thumbnail_size)
        scale = (max_log - min_log) / (
            self.horizontalSlider_icon_size.maximum() - self.horizontalSlider_icon_size.minimum()
        )
        linear_value = (
            math.log(value) - min_log
        ) / scale + self.horizontalSlider_icon_size.minimum()
        return int(linear_value)

    def init_image_list(self) -> None:
        """
        Populate the image list with the images.
        """
        # Reduce the size of the list. Make it a 1/3 split.
        window_width = self.width()
        self.splitter.setSizes([window_width // 3, 2 * window_width // 3])

        label_text = f_plural(len(self.images), self.tr("image"), self.tr("images"))
        self.label_image_count.setText(f"{len(self.images)} {label_text}")

        for image in self.images:

            original_path = image.path
            if self.editing_old_data:
                output_path = original_path
            else:
                output_path = image.outputs[self.target_output].path

            if output_path is None:
                logger.error(f"Output path for {image.path} is None.")

            try:
                self.image_list.addItem(
                    Qw.QListWidgetItem(Qg.QIcon(str(output_path)), str(original_path.stem))
                )
            except OSError:
                gu.show_exception(
                    self,
                    self.tr("Loading Error"),
                    self.tr("Failed to load image '{path}'").format(path=output_path),
                )

        # Use a logarithmic scale to distribute the values.
        self.horizontalSlider_icon_size.valueChanged.connect(self.update_icon_size)
        self.horizontalSlider_icon_size.setValue(self.from_log_scale(THUMBNAIL_SIZE))

        # Sanity check.
        if len(self.images) != self.image_list.count():
            logger.error("Failed to populate the image list correctly.")

    def init_arrow_buttons(self) -> None:
        """
        Connect the arrow buttons to the image list.
        """
        self.pushButton_next.clicked.connect(
            lambda: self.image_list.setCurrentRow(self.image_list.currentRow() + 1)
        )
        self.pushButton_prev.clicked.connect(
            lambda: self.image_list.setCurrentRow(self.image_list.currentRow() - 1)
        )

        # Disable the buttons if there is no next or previous image.
        self.image_list.currentRowChanged.connect(self.check_arrow_buttons)

    def check_arrow_buttons(self) -> None:
        """
        Check if the arrow buttons should be enabled or disabled.
        """
        current_row = self.image_list.currentRow()
        self.pushButton_prev.setEnabled(current_row > 0)
        self.pushButton_next.setEnabled(current_row < self.image_list.count() - 1)

    @Slot(int)
    def update_icon_size(self, size: int) -> None:
        """
        Update the icon size in the image list.

        :param size: The new size.
        """
        size = self.to_log_scale(size)
        self.image_list.setIconSize(Qc.QSize(size, size))
        for index, image in enumerate(self.images):
            item = self.image_list.item(index)
            text = image.path.name
            self.elide_text(item, text, int(size * 0.9))

    def elide_text(self, item, text, width):
        font_metrics = Qg.QFontMetrics(self.font())
        elided_text = font_metrics.elidedText(text, Qc.Qt.ElideLeft, width)
        item.setText(elided_text)

    @Slot(Qw.QListWidgetItem, Qw.QListWidgetItem)
    @Slot(Qw.QListWidgetItem)
    def handle_image_change(
        self, current: Qw.QListWidgetItem = None, previous: Qw.QListWidgetItem = None
    ) -> None:
        """
        Figure out what the new image is and display it.
        """
        # We don't need the previous image, so we ignore it.
        index = self.image_list.currentRow()
        image = self.images[index]
        self.switch_to_image(image)

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(
            self,
            self.splitter,
            self.horizontalSlider_icon_size,
        )


class WrapTextDelegate(Qw.QStyledItemDelegate):
    """
    Let table widget rows expand vertically to prevent line wrapping.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._table_widget = parent

    def paint(self, painter, option, index):
        # Set text wrapping for the cell
        option.textElideMode = Qc.Qt.ElideNone
        option.displayAlignment = Qc.Qt.AlignLeft | Qc.Qt.AlignVCenter
        option.widget.setWordWrap(True)
        super().paint(painter, option, index)

    def sizeHint(self, option, index):
        # Get the text from the index
        text = index.data(Qc.Qt.DisplayRole)
        if not text:
            return super().sizeHint(option, index)

        # Create a Qg.QFontMetrics object to calculate the size of the text
        font_metrics = Qg.QFontMetrics(option.font)

        # Calculate the width available for the text
        text_width = option.rect.width()

        # Calculate the height needed for the text to wrap within the given width
        text_height = font_metrics.boundingRect(
            0, 0, text_width, 0, Qc.Qt.TextWordWrap, text
        ).height()

        # Add an extra half line height.
        text_height += font_metrics.height()

        # Calculate the default size and add text height if necessary
        size = super().sizeHint(option, index)
        return Qc.QSize(size.width(), max(size.height(), text_height))



from tests.helpers import mock_file_path
import pcleaner.image_ops as ops
import time


def test_split_image():
    image_path = mock_file_path("long_strip.jpg")

    preferred_height = 2000
    tolerance = 1000

    start = time.time()
    splits = ops.calculate_best_splits(
        image_path,
        preferred_height,
        tolerance,
        split_strips=True,
        max_aspect_ratio=0.3,
    )
    assert True

    if True:
        return

    print(f"Time taken: {(time.time() - start) * 1000:.2f} ms")
    print("Splits:")
    print(splits)

    # Draw the splits.
    from PIL import Image, ImageDraw
    from math import ceil

    image = Image.open(image_path)
    draw = ImageDraw.Draw(image)
    height = image.height
    num_strips = ceil(height / preferred_height) - 1
    preferred_height = height // (num_strips + 1)
    for index, split in enumerate(splits, 1):
        initial_split = preferred_height * index
        draw.line(
            (0, initial_split - tolerance, image.width, initial_split - tolerance),
            fill="blue",
            width=2,
        )
        draw.line((0, split, image.width, split), fill="red", width=5)
        draw.line(
            (0, initial_split + tolerance, image.width, initial_split + tolerance),
            fill="blue",
            width=2,
        )
        draw.line(
            (0, initial_split, image.width, initial_split),
            fill="green",
            width=2,
        )

    image.show()



from enum import IntEnum

import PySide6.QtCore as Qc
import PySide6.QtGui as Qg
import PySide6.QtWidgets as Qw
from loguru import logger

import pcleaner.gui.state_saver as ss
import pcleaner.ocr.supported_languages as osl
from pcleaner.gui.ui_generated_files.ui_OCRLanguageSupport import Ui_OCRLanguageSupport
from pcleaner.helpers import tr
from pcleaner.ocr import ocr


class Column(IntEnum):
    CODE = 0
    LANGUAGE = 1
    MANGAOCR = 2
    TESSERACT = 3


class OCRLanguageSupport(Qw.QDialog, Ui_OCRLanguageSupport):
    """
    Display the current state of OCR language support.
    """

    state_saver: ss.StateSaver  # The state saver for the window.

    def __init__(
        self,
        parent=None,
    ):
        """
        Init the widget.

        :param parent: The parent widget.
        """
        logger.info(f"Opening the OCR Language Support dialog.")
        Qw.QDialog.__init__(self, parent)
        self.setupUi(self)

        self.populate_table()

        # Connect the buttons.
        self.pushButton_install_tesseract.clicked.connect(self.open_tesseract_installation)
        self.pushButton_install_tesseract_language_packs.clicked.connect(
            self.open_tesseract_langs_installation
        )

        self.state_saver = ss.StateSaver("ocr_language_support")
        self.init_state_saver()
        self.state_saver.restore()

    def closeEvent(self, event: Qg.QCloseEvent) -> None:
        """
        Called when the window is closed.
        """
        self.state_saver.save()
        event.accept()

    def init_state_saver(self) -> None:
        """
        Load the state from the state saver.
        """
        self.state_saver.register(
            self,
            self.lang_table,
        )

    def populate_table(self):
        # I tried to make the columns sortable, so that you could figure out what all
        # the languages are, that something like tesseract currently supports,
        # but item delegates were not only a pain in the ass, but didn't work either.
        # Trying to use a label with text and making said text invisible also wouldn't
        # work with the icons.

        manga_ocr_langs = ocr.MangaOcr.langs()
        tesseract_langs = ocr.TesseractOcr.langs()

        if tesseract_langs:
            self.pushButton_install_tesseract.hide()

        for code, lang in osl.language_code_name_sorted(
            include_detect=False, pin_important=True, translate=tr
        ):
            if code in (osl.LanguageCode.detect_box, osl.LanguageCode.detect_page):
                continue
            self.lang_table.appendRow(str(code), tr(lang))

            row = self.lang_table.rowCount() - 1
            if code in manga_ocr_langs:
                self.lang_table.setCellWidget(
                    row, Column.MANGAOCR, IconWidget(Qg.QIcon.fromTheme("dialog-ok"))
                )

            if code in tesseract_langs:
                self.lang_table.setCellWidget(
                    row, Column.TESSERACT, IconWidget(Qg.QIcon.fromTheme("dialog-ok"))
                )

        self.lang_table.resizeColumnsToContents()

    @staticmethod
    def open_tesseract_installation():
        """
        Open the online documentation in the default browser.
        """
        logger.debug("Opening online documentation for Tesseract installation.")
        Qg.QDesktopServices.openUrl(
            Qc.QUrl("https://github.com/VoxelCubes/PanelCleaner?tab=readme-ov-file#ocr")
        )

    @staticmethod
    def open_tesseract_langs_installation():
        """
        Open the online documentation in the default browser.
        """
        logger.debug("Opening online documentation for Tesseract language pack installation.")
        Qg.QDesktopServices.openUrl(
            Qc.QUrl("https://ocrmypdf.readthedocs.io/en/latest/languages.html")
        )


class IconWidget(Qw.QWidget):
    def __init__(self, icon, parent=None):
        super().__init__(parent)
        layout = Qw.QHBoxLayout()
        label = Qw.QLabel()
        label.setPixmap(icon.pixmap(22, 22))  # Adjust size as necessary
        layout.addWidget(label)
        layout.setAlignment(label, Qc.Qt.AlignCenter)  # Center the QLabel
        layout.setContentsMargins(0, 0, 0, 0)  # Remove margins
        self.setLayout(layout)
